# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the FSLoader class, described below."""


import os
import os.path
import sys

from ytranslate.catalog import Catalog
from ytranslate.loader import Loader

class FSLoader(Loader):

    """A file system loader of catalogs.

    This class represents a loader for YAML catalogs.  Catalogs
    are contained in files.  Directories are used to create a hierarchy
    of namespaces.  The root directory given in argument is the
    parent directory.  The directories contained in this root
    directory are recursively explored for catalog files.

    """

    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.catalogs = {}
        self.namespaces = {}

    def __repr__(self):
        return "<ytranslate.FSLoader (root={})>".format(repr(self.root_dir))

    def load(self):
        """Load the catalogs."""
        root_dir = self.root_dir
        len_root = len(root_dir.split(os.sep))
        for base, dirs, files in os.walk(self.root_dir):
            for file in files:
                if len(file) > 4 and file.endswith(".yml"):
                    fullname = os.path.join(base, file)
                    kwargs = {}

                    # If Python 3, enforce the encoding to 'utf-8'
                    if sys.version_info.major == 3:
                        kwargs["encoding"] = "utf-8"

                    try:
                        with open(fullname, "r", **kwargs) as file:
                            data = file.read()
                    except IOError as e:
                        raise ValueError("cannot load the {} file: " \
                                "{}".format(repr(fullname), e))

                    namespace = ".".join(fullname[:-4].split(
                            os.sep)[len_root + 1:])

                    parent = fullname.split(os.sep)[len_root]
                    if parent.endswith(".yml"):
                        parent = parent[:-4]

                    catalog = Catalog(fullname)
                    catalog.read_YAML(data)
                    if parent != namespace:
                        if parent not in self.catalogs:
                            self.catalogs[parent] = Catalog(parent)
                        parent = self.catalogs[parent]
                        parent.copy_from(catalog,
                                namespace=namespace)
                    else:
                        self.catalogs[namespace] = catalog
                    self.namespaces[namespace] = catalog

    def update_catalog(self, catalog, model, missing="???"):
        """Update the given catalog.

        The catalog specified as a model is used to fill the information
        out, if not provided in the first catalog.  This method can
        be used to create the first catalog or to update it.

        Return the number of updated messages.

        """
        nb = 0
        model = self.catalogs[model]
        if catalog not in self.catalogs:
            catalog = Catalog(catalog)
            self.catalogs[catalog.name] = catalog
        else:
            catalog = self.catalogs[catalog]

        # Write the catalog with missing information
        for key, value in model.messages.items():
            replace = missing
            if isinstance(value, dict):
                replace = value.copy()
                for nkey in replace.keys():
                    replace[nkey] = missing

            if key not in catalog.messages:
                nb += 1
                catalog.messages[key] = replace

        # Finally, write the updated (or newly-created) catalog
        self.save_catalog(catalog)

        return nb

    def save(self):
        """Save all catalogs in the file system."""
        for namespace, catalog in self.namespaces.items():
            self.save_catalog(catalog)

    def save_catalog(self, catalog):
        """Save the specified catalog in the file system.

        Each catalog stores, in its name, the full path leading to
        it.  Assuming the current directory has remained the same
        and the directory structure hasn't significantly changed,
        this method should be able to access the proper
        file and write into it.  An IOError exception is bound to
        be raised if things didn't work for some reason.

        """
        name = catalog.name
        parent_directory = os.path.join(self.root_dir, name)
        kwargs = {}
        # If Python 3, enforce the encoding to 'utf-8'
        if sys.version_info.major == 3:
            kwargs["encoding"] = "utf-8"

        for namespace in self.namespaces.keys():
            if namespace:
                fullname = os.path.join(parent_directory,
                        namespace.replace(".", os.path.sep) + ".yml")
            else:
                parent_directory = self.root_dir
                fullname = os.path.join(parent_directory, name + ".yml")

            # Create the directory structure if necessary
            parent = os.path.split(fullname)[0]
            if not os.path.exists(parent):
                os.makedirs(parent)

            yaml = catalog.write_YAML(namespace)
            with open(fullname, "w", **kwargs) as file:
                file.write(yaml)



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the Loader class, described below."""


class Loader(object):

    """A base class for a catalog loader.

    Loaders should inherit this class, and implement the 'load'
    method.  A loader is used to define a way to load catalogs.
    For instance, FSLoader (File System Loader) loads catalogs from
    the file system.  If you wish to create a new type of loader,
    you can simply inherit this class.  A single loader is used
    by default (see the 'init' funciton in 'ytranslate.tools'),
    although it's possible to create several loaders and work with
    them simultaneously, if the need arises.

    """

    current_loader = None
    current_catalog = None

    def load(self):
        """Load the catalogs."""
        raise NotImplementedError

    def select(self, catalog):
        """Select the catalog of the specified name."""
        type(self).current_catalog = self.catalogs[catalog]



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import io
import os
import sys
from textwrap import dedent
import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from ytranslate.fsloader import FSLoader
from ytranslate.fsloader import os as fs_os

class TestFSLoader(unittest.TestCase):

    """Unittest for the FSLoader loader.

    This set of tests checks the features of the File System Loader
    (FSLoader) which is to loads the catalogs on the file system.
    In order to avoid testing the file system itself, access to it
    are 'mocked' by the 'mock' module ('unittest.mock' under Python
    3.3 and up).  You should install the 'mock' module with
    'pip install mock' if you're running Python 2.X.

    """

    def setUp(self):
        self.files = {}

    def open(self):
        """Return a mock of the open function."""
        address_open = "builtins.open"
        if sys.version_info.major == 2:
            address_open = "ytranslate.fsloader.open"

        return mock.patch(address_open, mock.mock_open())

    def mock_open(self, name, *args, **kwargs):
        """Open a mock file at the proper location.

        The match between file name and content is done in the 'files'
        dictionary.  A 'io.StringIO' object is returned.  This
        function serves as a substitute for 'open' in a test context.

        """
        name = name.replace(os.sep, "/")
        content = self.files.get(name, "")
        return io.StringIO(u"{}".format(content))

    @mock.patch.object(fs_os, "walk")
    def test_files(self, mock_walk):
        """Simple test with only files.

        The purpose of this test is to ensure that first-level
        catalogs are corectly interpreted.  These catalogs are files
        directly in the given directory.  Their name becomes their
        namespace without the '.yml' extension.  Example: 'en.yml'
        is stored in the 'en' namespace.

        """
        # Initialize the files
        self.files = {
                "test/en.yml": dedent("""\
                        new: New
                        view: View"""),
                "test/fr.yml": dedent("""\
                        new: Nouveau
                        view: Affichage"""),
        }

        with self.open() as mock_open:
            mock_open.side_effect = self.mock_open
            mock_walk.return_value = [
                ["test", [], ["en.yml", "fr.yml"]],
            ]

            # Creates the FSLoader object
            loader = FSLoader("unknown")
            loader.load()

            # Check the catalogs' proper loading
            self.assertEqual(len(loader.catalogs), 2)
            self.assertIn("en", loader.catalogs)
            self.assertIn("fr", loader.catalogs)

            # Check the catalog's content
            en = loader.catalogs["en"]
            fr = loader.catalogs["fr"]
            self.assertEqual(en.retrieve("new"), u"New")
            self.assertEqual(fr.retrieve("new"), u"Nouveau")
            self.assertEqual(en.retrieve("view"), u"View")
            self.assertEqual(fr.retrieve("view"), u"Affichage")

    @mock.patch.object(fs_os, "walk")
    def test_directories(self, mock_walk):
        """Test with files and directories.

        This test checks the interpretation of second-level catalogs,
        that is, when a full structure of directories is used to
        represent namespaces.  In this case, the namespaces are
        the names of directories excluding the first one, which
        represents the main namespace.  For instance,
        here's a possible directory structure:
            en/
                ui/
                    window.yml
                    errors.yml
                    ...

        """
        # Initialize the files
        directories = [
            ["example", ["en", "fr"], []],
            ["example/en", ["ui", "message"], []],
            ["example/en/ui", [], ["window.yml", "errors.yml"]],
            ["example/en/message", [], ["notification.yml"]],
            ["example/fr", ["ui", "message"], []],
            ["example/fr/ui", [], ["window.yml", "errors.yml"]],
            ["example/fr/message", [], ["notification.yml"]],
        ]

        for i, elt in enumerate(directories):
            directories[i][0] = elt[0].replace("/", os.sep)

        self.files = {
                "example/en/ui/window.yml": dedent("""\
                        title: Ytranslator
                        buttons:
                            exit: Exit
                            quit: Quit"""),
                "example/en/ui/errors.yml": dedent("""\
                        syntax: syntax error"""),
                "example/en/message/notification.yml": dedent("""\
                        email: You have some unread e-mails"""),
                "example/fr/ui/window.yml": dedent("""\
                        title: Ytraducteur
                        buttons:
                            exit: Fermer
                            quit: Quitter"""),
                "example/fr/ui/errors.yml": dedent("""\
                        syntax: erreur de syntaxe"""),
                "example/fr/message/notification.yml": dedent("""\
                        email: Vous avez des messages non lus"""),
        }

        with self.open() as mock_open:
            mock_open.side_effect = self.mock_open
            mock_walk.return_value = directories

            # Creates the FSLoader object
            loader = FSLoader("unknown")
            loader.load()

            # Check the catalogs' proper loading
            self.assertEqual(len(loader.catalogs), 2)
            self.assertIn("en", loader.catalogs)
            self.assertIn("fr", loader.catalogs)

            # Check the catalog's content
            en = loader.catalogs["en"]
            fr = loader.catalogs["fr"]
            self.assertEqual(en.retrieve("ui.window.title"), u"Ytranslator")
            self.assertEqual(en.retrieve("ui.window.buttons.quit"), u"Quit")
            self.assertEqual(en.retrieve("ui.errors.syntax"), u"syntax error")
            self.assertEqual(fr.retrieve("ui.window.title"), u"Ytraducteur")
            self.assertEqual(fr.retrieve("ui.window.buttons.quit"), u"Quitter")
            self.assertEqual(fr.retrieve("ui.errors.syntax"),
                    u"erreur de syntaxe")



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing the Catalog class, described below."""

import yaml

try:
    unicode
except NameError:
    unicode = str

class Catalog:

    """A losely-defined catalog for translation.

    A Catalog object contains a hierarchy of matches and sub-catalogs.
    This hierarchy, established by the developer in charge of
    the translations, must be reproduced for every language that
    the software should support.  These catalogs are defined as
    YML files on the file system, although they may be created and
    expanded programmatically.  Also note that catalogs can easily
    be used in other situations, when simple and complex matching applies.

    A simple catalog defined in YAML, en.yml, may look like this:
        ui:
            title: Ytranslator
            login:
                unknown: Welcome.  Perhaps you would wish to login?
                user: Welcome, {user}!  Glad to see you back.

    In this example, the catalog 'en' contains a single namespace:
    'ui'.  In it are a simple message, 'title', and anther namespace,
    'login', which contains two messages.  You could use this catalog
    in your code like this:
        from ytranslate import select, t
        # Select the 'en' catalog that we just defined
        select("en")
        # Retrive the title of the GUI application
        title = t("ui.title")
        # Is the user logged in?
        if user.logged_in:
            print t("ui.login.user", name="Jean")
        else:
            print t("ui.login.unknown")

    Of course, having a single catalog doesn't make much sense.  But
    if you have a catalog in a different language using exactly the
    same hierarchy, things look very interesting.  You could have
    a 'fr.yml' catalog for instance:
        ui:
            title: Ytraducteur
            login:
                unknown: Bienvenue. Souhaiteriez-vous vous connecter ?
                user: Bienvenue, {user} !  Heureux de vous revoir.

    You can try to run the same code selecting 'fr' for a catalog
    this time.  The concept is pretty simple, but the possibilities
    are endless.  Most catalogs of complex applications are defined
    in folders and sub-folers, which are just new namespaces in
    the hierarchy.

    """

    def __init__(self, name):
        self.name = name
        self.messages = {}

    def __repr__(self):
        return "<ytranslate.Catalog {}>".format(repr(self.name))

    def read_dictionary(self, dictionary, parent=""):
        """Read a namespace defined in a dictionary."""
        for name, entry in dictionary.items():
            name = str(name)
            if parent:
                name = parent + "." + name

            if isinstance(entry, dict):
                if all(str(key).rstrip("+").isdigit() for key in entry):
                    copied = {}
                    for key, value in entry.items():
                        copied[str(key)] = unicode(value)

                    self.messages[name] = copied
                else:
                    self.read_dictionary(entry, parent=name)
            else:
                self.messages[name] = unicode(entry)

    def read_YAML(self, content):
        """Fill the catalog using this YAML content.

        The content must be a str containing YAML code that should
        represent a dictionary.  The whole dictionary will be integrated
        in the 'messages' dictionary.  Namespaces can be deefined
        in the YAML document, but they will be "flattened out".
        For instance, if the content contains the following string:
            file: Fichier
            edit: Édition
            view: Affichage
            connection:
                connected: Connecté
                connecting: Connexion en cours
                error: Connexion impossible

        Then the following messages will be created or updated:
        {
            'file': 'Fichier',
            'edit': 'Édition',
            'view': 'Affichage',
            'connection.connected': 'Connecté',
            'connection.connecting': 'Connexion en cours',
            'connection.error': 'Connexion impossible',
        }

        """
        try:
            data = yaml.safe_load(content)
        except yaml.parser.ParserError as err:
            # Use a newly-defined expression
            raise ValueError("an error occurred while parsing the " \
                    "YAML content:\n{}".format(str(err)))

        if isinstance(data, dict):
            self.read_dictionary(data)
        else:
            raise ValueError("catalog {}, the YAML content doesn't " \
                    "describe a dictionary".format(self.name))

    def copy_from(self, catalog, namespace=""):
        """Copy the messages of the catalog provided as a parameter.

        The 'namespace' parameter can create sub-namespaces for
        a whole catalog.

        """
        for name, message in catalog.messages.items():
            if namespace:
                name = namespace + "." + name

            self.messages[name] = message

    def write_dictionary(self, root=""):
        """Write the nested dictionary.

        Namespaces are used to create a nested dictionary.  The
        'ui.title' message would be converted in a dictionary with
        a key 'ui', and as a value another dictionary containing
        in key 'title' and in value the message's value.  The
        process is exactly the opposite of the 'read_dictionary'
        method.

        If a namespace is specified as a root, only select this namespace
        in the list of keys, returning the dictionary as if it were
        a root.

        """
        keys = sorted(self.messages.keys())
        if root:
            keys = [key for key in keys if key.startswith(root + ".")]

        nested = {}
        for key in keys:
            value = self.messages[key]
            if root:
                key = key[len(root) + 1:]

            # Split the key in namespaces separated by '.'
            last_namespace = key.split(".")[-1]
            namespaces = key.split(".")[:-1]
            current = nested
            for namespace in namespaces:
                if namespace not in current:
                    current[namespace] = {}

                current = current[namespace]

            current[last_namespace] = value

        return nested

    def write_YAML(self, root=""):
        """Return the nested content as YAML."""
        nested = self.write_dictionary(root)
        return yaml.safe_dump(nested, indent=4, width=79,
                default_flow_style=False)

    def retrieve(self, address, count=None, **kwargs):
        """Retrieve the message using the specified address.

        The address is a string of namespaces separated by '.'.
        For instance, "ui.greeting.user".  This address must match
        both the directory/file structure and the internal catalog
        structure.

        It is possible to specify placeholders in the message.  For
        instance, let's say you have a message defined like this
        in your YAML catalog:
            greeting: Hello, {name}!

        In the call to 'retrieve', you will be able to specify
        the name:
            message = catalog.retrieve("greeting", name="John")
            # message will contain "Hello, John!"

        You can also alter the message depending on a number
        (singular and plural in English, or a more specific variation
        in Russian, for instance).  The full syntax is described
        in the 'retrieve_count' method.

        """
        message = self.messages.get(address)
        if count is None and message is None:
            raise ValueError("address {} cannot be found in this " \
                    "catalog".format(repr(address)))

        if count is not None:
            message = self.retrieve_count(address, count, **kwargs)

        return message.format(count=count, **kwargs)

    def retrieve_count(self, address, count, **kwargs):
        """Retrieve a message when a 'count' indicator is present.

        The 'retrieve' method calls 'retrieve_count' if necessary.
        This method shouldn't be called directly.

        A 'count' indicator can be used to alter the message regarding singular, plural or other rules that may apply to the language depending on a number of items.  Definition a group of message in a catalog is simple:  A new dictionary has to be specified instead of just a message.  For instance:
            emails:
                0: You have no email.
                1: You have one email.
                2+: You have {count} emails.

        Then you can simply ask the 'retrieve' method to find the proper message according to the 'count' indicator:
            t("emails", 2)
            # Which will retrieve the message 'You have 2 emails.'

        This second dictionary defined in YAML must contain only
        numbers as keys, or a number followed by the '+' sign,
        which means "either that number of more".  For instance,
        '5+' means either 5 or more.  This way, you can configure
        the catalog to different singular and plural rules
        depending on the language.

        """
        equal = lambda a, b: a == b
        greater = lambda a, b: a >= b
        messages = self.messages.get(address, {})
        if messages is None:
            raise ValueError("address {} cannot be found in this " \
                    "catalog".format(repr(address)))
        elif not isinstance(messages, dict):
            raise ValueError("the message at {} has to be " \
                    "retrieved with a 'count' indicator, though " \
                    "at this address aren't several values".format(
                    repr(address)))

        message = None
        for key in reversed(list(messages.keys())):
            value = messages[key]
            compare = equal
            if key.endswith("+"):
                key = key[:-1]
                compare = greater

            if not key.isdigit():
                raise ValueError("message {}: {} isn't a valid " \
                        "number.".format(repr(address), key))

            key = int(key)
            if compare(count, key):
                message = value
                break

        if message is None:
            raise ValueError("address {}: no proper message " \
                    "to be displayed with a count of {}".format(
                    repr(address), count))

        return message



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing different tools as functions.

They can be called to interact with the created object of the
'ytranslate' library, including the catalogs and loader.

"""

from ytranslate.fsloader import FSLoader
from ytranslate.loader import Loader

def init(LoaderClass=FSLoader, **kwargs):
    """Load the catalogs at a specified location.

    Depending on the type of loader, the arguments vary.  If the
    default LoaderClass is used, the first argument is the root directory.
    The 'root_dir', the parent directory, is sent to the FSLoader
    class which is to create a hierarchy of catalogs.  The parent
    catalogs bear the name of the namespace (that is their
    directory or their filename without the '.yml' extension).
    For instance:
        init(root_dir="path/to/translations")

    Use the 'select' function to then select a catalog.

    """
    loader = LoaderClass(**kwargs)
    FSLoader.current_loader = loader
    loader.load()

def select(catalog):
    """Select the catalog from the loader.

    The catalog's name must be specified.  If the loader is a
    FSLoader (the default), then the 'root_dir' directory contains
    the parent catalogs.  You should use one of its contained
    directoriess' names, or that of a MYL file without the '.yml'
    extension.  For instance:
        select("en")

    """
    if FSLoader.current_loader:
        FSLoader.current_loader.select(catalog)
    else:
        raise ValueError("the current loader hasn't been selected")

def t(address, count=None, **kwargs):
    """Retrieve the translated message from the selected catalog.

    You can use this function to obtain the translated message,
    corresponding to the address, which must represent the list of
    namespaces separated by '.'.  For instance:
        t("ui.title")

    The hierarchy of messages is defined by the catalog's structure
    (directories and files, if it has been selected by a FSLoader,
    which is the default choice).

    You can also use placeholders as named parameters:
        t("welcome.name", user="John")

    Additionally, you can vary the message according to a number.
    For instance:
        t("notificaiton.emails", 3)

    See the user documentation for a detailed explanation about
    the syntax and corresponding catalogs.

    """
    if FSLoader.current_catalog:
        return FSLoader.current_catalog.retrieve(address, count, **kwargs)

    raise ValueError("no catalog has been selected")



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Module containing the UpdateCommand class, described below."""

from __future__ import print_function
import os
import os.path
import sys

from ytranslate.commands.base import BaseCommand
from ytranslate.fsloader import FSLoader

class UpdateCommand(BaseCommand):

    """Commands 'update'.

    This command updates a given catalog with a default model.

    """

    name = "update"

    def __init__(self, parser=None):
        BaseCommand.__init__(self, parser)
        parser.add_argument("directory",
                help="the path to the directory containing the catalogs")
        parser.add_argument("catalog", nargs='?',
                help="the catalog name to be created or updated")
        parser.add_argument("-m", "--model", nargs='?',
                default="en", help="the model catalog to be used")

    def execute(self, args):
        """Execute the command."""
        root_dir = args.directory
        if not os.path.exists(root_dir):
            print("The {} directory doesn't exist".format(repr(root_dir)),
                    file=sys.stderr)
            sys.exit(1)
        elif not os.path.isdir(root_dir):
            print("The {} path doesn't lead to a directory".format(
                    repr(root_dir)), file=sys.stderr)
            sys.exit(1)

        loader = FSLoader(root_dir)
        loader.load()
        nb = loader.update_catalog(args.catalog, args.model)
        print("Successfully updated the '{}' catalog ({})".format(
                args.catalog, nb))



# Copyright (c) 2015, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

from ytranslate.catalog import Catalog

# Documents
SIMPLE_DOC = """
file: Fichier
edit: Édition
view: Affichage
greeting: Bienvenue, {name} !
connection:
    connected: Connecté
    connecting: Connexion en cours
    error: Connexion impossible
""".strip()

PLURAL_DOC = """
emails:
    0: You have no email
    1: Well, you have one email
    2+: You only have {count} emails
    5+: Wow, you have {count} emails
""".strip()

class TestCatalog(unittest.TestCase):

    """Unittest for the Catalog class.

    This set of unittests is to ensure the proper building of
    catalogs using YAML documents.

    """

    def test_simple_structure(self):
        """Test to read and create the proper structure."""
        catalog = Catalog("test")
        catalog.read_YAML(SIMPLE_DOC)
        self.assertEqual(catalog.messages, {
            'file': u'Fichier',
            'edit': u'Édition',
            'view': u'Affichage',
            'greeting': 'Bienvenue, {name} !',
            'connection.connected': u'Connecté',
            'connection.connecting': u'Connexion en cours',
            'connection.error': u'Connexion impossible',
        })

    def test_write_dictionary(self):
        """Test to read and retrive the nested structure."""
        catalog = Catalog("test")
        catalog.read_YAML(SIMPLE_DOC)
        dictionary = catalog.write_dictionary()
        self.assertEqual(dictionary, {
            'file': u'Fichier',
            'edit': u'Édition',
            'view': u'Affichage',
            'greeting': 'Bienvenue, {name} !',
            'connection': {
                'connected': u'Connecté',
                'connecting': u'Connexion en cours',
                'error': u'Connexion impossible',
            },
        })

    def test_retrieve(self):
        """Test to retrieve the proper message from a catalog."""
        catalog = Catalog("test")
        catalog.read_YAML(SIMPLE_DOC)
        self.assertEqual(catalog.retrieve("view"), "Affichage")
        self.assertEqual(catalog.retrieve("connection.error"),
                "Connexion impossible")
        self.assertEqual(catalog.retrieve("greeting", name="Jeanne"),
                "Bienvenue, Jeanne !")

    def test_retrieve_plural(self):
        """Test to retrieve the proper message using count indicators."""
        catalog = Catalog("test")
        catalog.read_YAML(PLURAL_DOC)
        self.assertEqual(catalog.retrieve("emails", 0), "You have no email")
        self.assertEqual(catalog.retrieve("emails", 1),
                "Well, you have one email")
        self.assertEqual(catalog.retrieve("emails", 2),
                "You only have 2 emails")
        self.assertEqual(catalog.retrieve("emails", 3),
                "You only have 3 emails")
        self.assertEqual(catalog.retrieve("emails", 5),
                "Wow, you have 5 emails")
        self.assertEqual(catalog.retrieve("emails", 6),
                "Wow, you have 6 emails")



from setuptools import setup, find_packages

DESCRIPTION = """
Ytranslate (Why translate) is a simple tool to manage translations
through creating and updating losely-defined YAML catalogs.
""".strip()

setup(
    name = "ytranslate",
    version = "0.3",
    packages = find_packages(),
    install_requires = ['pyyaml>=3'],
    description = DESCRIPTION,
    author = 'Vincent Le Goff',
    author_email = 'vincent.legoff.srs@gmail.com',
    url = 'https://github.com/vlegoff/ytranslate',
    keywords = ['translation', 'translate', 'yaml', 'yml'],
    entry_points = {
        'console_scripts': ['ytranslate=ytranslate.commands.main:main'],
    },
)